// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 1998-2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: CoveringRelations.cc,v 1.2 2002/06/12 22:33:36 carzanig Exp $
//

#include <siena/Siena.h>

#include "CoveringRelations.h"

#ifdef NO_INLINE
#define inline
#include "CoveringRelations.icc"
#endif

//
// semantics of subscriptions
//
bool covers(const Filter &f, const Event &e)
{
    //
    // covers(f,e) <==> for each attribute filter af in f, there
    // exists an attribute ae in e such that: ae.name == af.name &&
    // apply_operator(af.operator, ae.value, af.value)
    //
    // NOTE: an empty filter (no constraints) yields always false!
    //
    Filter::const_iterator	fi;
    Event::const_iterator	ei;

    ei = e.begin();
    fi = f.begin();

    if (fi == f.end()) return false;
    do {
	ei = e.find((*fi).first);
	if (ei == e.end() 
	    || !apply_operator((*fi).second.op, (*ei).second, (*fi).second))
	    return false;
    } while (++fi != f.end());
    return true;
}

bool covers(const AttributeConstraint &af1, const AttributeConstraint &af2)
{
    //
    // true when af1 defines a set of events S_af1 that contains the
    // set of events S_af2 defined by af2, in other words true iff af2
    // ==> af1 , i.e., for every Attribute x: x op2 f2 ==> x op1 f1
    // where op2 is the operator defined by af2 and f2 is the value
    // defined by af2. Same thing for af1.
    //
    // this is a crucial function! it is also quite tricky, think
    // twice before you change this implementation
    //
    // All Siena operators define transitive relations, except for
    // Siena_ne.
    //
    // trivial cases
    //
    // {x any} covers everything
    //
    if (af1.op == Siena_xx) return true;
    //
    // nothing covers {x any} (except {x any}, see above)
    //
    if (af2.op == Siena_xx) return false;
    //
    // {x != a} C {x op b} <-- not a op b
    //
    if (af1.op == Siena_ne) 
	return !apply_operator(af2.op,af1, af2);
    //
    // same operators (we already excluded af1.op == Siena_ne)
    //
    if (af2.op == af1.op)
	return apply_operator(af1.op, af2, af1) 
	    || apply_operator(Siena_eq, af2, af1);
    //
    // else I must consider the implications between DIFFERENT operators
    //
    switch(af2.op) {
    case Siena_eq: return apply_operator(af1.op, af2, af1);
    case Siena_lt: return af1.op == Siena_le 
		       && apply_operator(Siena_lt,af2,af1);
    case Siena_le: return af1.op == Siena_lt
		       && apply_operator(Siena_gt,af1,af2);
    case Siena_gt: return af1.op == Siena_ge
		       && apply_operator(Siena_le,af1,af2);
    case Siena_ge: return af1.op == Siena_gt
		       && apply_operator(Siena_lt,af1,af2);
    case Siena_sf: return af1.op == Siena_ss
		       && apply_operator(Siena_ss,af2,af1);
    case Siena_pf: 
	switch(af1.op) {
	case Siena_ss: return apply_operator(Siena_ss,af2,af1);
	case Siena_gt: return apply_operator(Siena_le,af1,af2)
			   && !apply_operator(Siena_pf,af1,af2);
	case Siena_lt: return apply_operator(Siena_ge,af1,af2)
			   && !apply_operator(Siena_pf,af1,af2);
	case Siena_ge: return apply_operator(Siena_lt,af1,af2);
	case Siena_le: return apply_operator(Siena_gt,af1,af2);
	default: return false;
	}
    default: return false;
    }
}

bool negates(const AttributeConstraint &af1, const AttributeConstraint &af2)
{
    // ...work in progress... DO NOT USE!
    //
    // true when 
    // \foreach value x 
    //   apply_operator(af1.op, x, af1) ==> !apply_operator(af2.op, x, af2)
    // 

    switch(af1.op) {
    case Siena_eq: return !apply_operator(af2.op, af1, af2);
    case Siena_lt: return (af2.op == Siena_gt || af2.op == Siena_ge)
		       && apply_operator(Siena_ge,af2,af1);
    case Siena_le: return (af2.op == Siena_gt || af2.op == Siena_ge)
		       && !apply_operator(af2.op,af1,af2);
    case Siena_pf: return af2.op == Siena_pf
		       && !apply_operator(Siena_pf, af1, af2) 
		       && !apply_operator(Siena_pf, af2, af1);
    case Siena_sf: return !apply_operator(Siena_sf, af1, af2) 
		       && !apply_operator(Siena_sf, af2, af1);
    default: return false;
    }
}

bool covers(const Filter &f1, const Filter &f2)
{
    //
    // true iff f2 ==> f1  
    // I think this expression translates into:
    // for each attribute filter af1 in f1, there exist at least one
    // corresponding (same name) attribute filter af2 in f2 such that 
    // af1 covers af2
    //
    // I'm not 100% sure of the demonstration though, the idea is that
    // attribute filters define ``connected'' subsets of ``ordered''
    // sets (which has changed since I added Siena_ne)... I think I
    // should also assume that f2 is not `null' (i.e., contradictory)
    // ...work in progress...
    //
    Filter::const_iterator		fi1;
    Filter::const_iterator		fi2;
    pair<Filter::const_iterator, Filter::const_iterator> range;

    for(fi1 = f1.begin(); fi1 != f1.end(); ++fi1) {
	bool found = false;
	range = f2.equal_range((*fi1).first);
	for(fi2 = range.first; !found && fi2 != range.second; ++fi2)
	    if (covers((*fi1).second, (*fi2).second))
		found = true;
	if (!found) return false;
    }
    return true;
}

Filter & diff(Filter & d, const Filter & f1, const Filter & f2)
{
    // f1 && diff(f1,f2) <==> f2
    // 
    // it is probably necessary to assume that f1 covers f2, i.e.:
    //
    // (f2 ==> f1) && f1 && diff(f1,f2) <==> f2
    //
    // the idea is to use the algorithm that computes the covering
    // relation covers(f2,f1). When the algorithm doesn't find an
    // attribute filter af2 that covers the corresponding af1, it
    // simply adds it to the difference.
    //
    Filter::const_iterator		fi1;
    Filter::const_iterator		fi2;
    pair<Filter::const_iterator, Filter::const_iterator> range;

    d.clear();
    for(fi2 = f2.begin(); fi2 != f2.end(); ++fi2) {
	bool found = false;
	range = f1.equal_range((*fi2).first);
	for(fi1 = range.first; !found && fi1 != range.second; ++fi1)
	    if (covers((*fi2).second, (*fi1).second))
		found = true;
	if (!found) {
	    range = d.equal_range((*fi2).first);
	    for(fi1 = range.first; !found && fi1 != range.second; ++fi1)
		if (covers((*fi2).second, (*fi1).second))
		    found = true;
	    d.insert(Filter::value_type((*fi2).first,(*fi2).second));
	}
    }
    return d;
}

Filter & simplify(Filter &f)
{
    //
    // this function eliminates implied constraints, e.g., if:
    // f = { x > 10, x < 30, x < 20, y > 0, y > 2 }
    // then, the result (i.e., f) will be:
    // f =  { x > 10, x < 20, y > 2 }
    //
    Filter::iterator current;
    Filter::iterator cursor;
    current = f.begin(); 
    while(current != f.end()) {
	cursor = current;
	++cursor;
	while(cursor != f.end() && (*cursor).first == (*current).first) {
	    if (covers((*current).second, (*cursor).second)) {
		f.erase(current++);
		break;
	    } else if (covers((*cursor).second, (*current).second)) {
		f.erase(current++);
	    } else {
		++current;
	    }
	}
    }
    return f;
}

Filter & intersection(Filter & f1, const Filter & f2)
{
    //
    // intersection of two filters:
    //
    // intersection(f1, f2) <==> f1 && f2
    //
    // this is probably not the best implementation ...work in progress...
    //  
    Filter::const_iterator i;
    for(i = f2.begin(); i != f2.end(); ++i) f1.insert(*i);
    return simplify(f1);
}

//  Filter::iterator Filter::insert(const Filter::value_type &x) 
//  {
//      iterator i;

//      for(i = lower_bound(x.first); i != upper_bound(x.first); ++i)
//  	if (covers_afilter_afilter(x.second, (*i).second)) {
//  	    (*i).second = x.second;
//  	    return i;
//  	} else if (covers_afilter_afilter(x.second, (*i).second)) {
//  	    return i;
//  	} else if (negate_afilter_afilter((*i).second, x.second)) {
//  	    throw(NullFilter(x.first, (*i).second, x.second));
//  	}
//      return insert(i,x);
//  }

