// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 1998-2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: EventServer.h,v 1.3 2002/11/22 20:57:45 carzanig Exp $
//
#ifndef _EventServer_h
#define _EventServer_h

#include <iostream>
#include <set>
#include <map>

#include "siena/Siena.h"
#include "siena/URI.h"
#include "siena/Monitor.h"

#include "Db.h"

#ifdef HAVE_CXX_NAMESPACE
using namespace Siena;
#endif

class EventServer;

class EventServer {
    public:
    static const string Version;
    //
    // This is the interface of the Event Server
    // these functions should not throw exceptions, i.e., they should
    // deal with exceptions themselves.
    //
    void		publish(Request &r, Event &e);
    void		subscribe(Request &r, Filter &f);
    void		unsubscribe(Request &r, Filter &f);
    void		advertise(Request &r, Filter &f);
    void		unadvertise(Request &r, Filter &f);
    void		connect_peer(Request &);
    void		unsubscribe_all(Request &);

    void		suspend(Request &r);
    void		resume(Request &r);

    void		shutdown();
    //
    // initialize takes the URI of the first peer connection as a
    // parameter, I'm not sure this is the right thing to do. I will
    // probably change this.
    //
    void		initialize(const string &my_uri,
				   const string &peer_uri);
    void		initialize(const string &my_uri);
    void		set_db(const char *filename);
    void		set_log(ostream *l);
    void		set_monitor(Monitor *m);
    unsigned short	attempts();
    unsigned short	attempts(unsigned short ma);

    //
    // this is for debugging purposes (not really part of the interface)
    //
    void		print_poset(ostream &os);

    EventServer(): root(Node(Filter())), db(NULL), cleaning_pending(false),
    max_attempts(3) {};

    private:
    //
    // this is the physical representation of the subscription poset:
    // the main entity is the Node class.
    //
    class Node {
    public:
	unsigned long		id;
	set<URI>		subscribers;
	const Filter		filter;
	set<Node*>		preset;
	set<Node*>		postset;

	Node(const Filter &f): 
	    subscribers(), filter(f), preset(), postset() {};
    };
    //
    // event server data:
    // 
    ostream *		logf;
    Monitor *		monitor;
    string		myuri;
    set<URI>		neighbors;
    map<URI, u_short>	pending;
    set<URI>		suspended;
    Node		root;
    Db			*db;
    bool		cleaning_pending;
    unsigned short	max_attempts;

    static const int	DefaultTTL = 30;
    //
    // ``server'' operations (mainly SENP)
    //
    void write_to_siena(const URI &u, Request &r);
    void write_to_siena(const URI &u, Request &r, const Event &e);
    void write_to_siena(const URI &u, Request &r, const Filter &e);

    Request & massage_request(Request &r);

    //
    // subscriptions table and misc operations
    //
    bool exists_in_predecessors(list<Node*> &, const Filter &, const URI &);
    void successors(list<Node*> &, const Filter &);
    void predecessors(list<Node*> &, const Filter &);
    void connect_nodes(Node *, Node *);
    void disconnect_nodes(Node *, Node *);
    void insert_node_between(Node *, list<Node*> &, list<Node*> &);
    void erase_node(Node * node);
    bool list_successors(set<Node*> & succ, const Filter &f);
    void clear_subtree(const Node *, const URI &);
    bool is_root(const Node *n);
    URI single_subscriber(const Node *n);
    void forwards(set<URI> &res, const Node *n);
    void unsubscribe_from_succ(Request & r, list<Node*> & succ, const URI & U);
    bool is_indirect_successor(const Node *x, const Node *y);
    void db_erase_node(const Node *node);
    void db_save_node(const Node *node);
    void db_save_neighbors();
    void clean_up_pending();
};

#endif
